% Show several pMRI reconstruction techniques 
%
% (c) Matthieu Guerquin-Kern, 2007.11.15


%% Simple Example
clear all;
close all;
%x = LoadSignal('MRI', 256);
x = double(imread('genou.jpg'));

S1 = zeros(size(x));
S2 = zeros(size(x));
S1(1:end/2,:)=1;
S2(end/2+1:end,:)=1;

x1 = S1.*x;
x2 = S2.*x;

X1 = fftn(fftshift(x1));
X2 = fftn(fftshift(x2));

Z1 = X1(1:2:end,:);
z1 = ifftshift(real(ifftn(Z1)));
Z2 = X2(1:2:end,:);
z2 = ifftshift(real(ifftn(Z2)));

z = [fftshift(z1,1);fftshift(z2,1)];

figure(1);
subplot(2,2,1);
imshow(uint8(x));
xlabel('orignal');
subplot(2,2,3);
imshow(uint8(z1));
xlabel('reduced FOV coil 1');
subplot(2,2,4);
imshow(uint8(z2));
xlabel('reduced FOV coil 2');
subplot(2,2,2);
imshow(uint8(z));
xlabel('reconstruction');
title('Simple example of pMRI reconstruction');

%% PILS
clear all;
close all;
%x = LoadSignal('MRI', 256);
x = double(imread('genou.jpg'));

Ncoils = 4;
R = 2;
[Ly,Lx] = size(x);
Ls = 1.5*Ly;
S = 1/2-1/2*cos(2*Ncoils*pi*(0:(Ls/Ncoils-1))/Ls)'*ones(1,Lx);

S1=zeros(size(x));
tmp = S(end/6+1:end,:);
S1(1:size(tmp,1),:) = tmp;
S2=zeros(size(x));
S2(Ly/Ncoils+1-size(S,1)/6:Ly/Ncoils+size(S,1)*5/6,:) = S;
S3=zeros(size(x));
S3(2*Ly/Ncoils+1-size(S,1)/6:2*Ly/Ncoils+size(S,1)*5/6,:) = S;
S4=zeros(size(x));
tmp = S(1:end*5/6,:);
S4(3*Ly/Ncoils+1-size(S,1)/6:end,:) = tmp;

Stot=S1+S2+S3+S4;

X1 = fftn(x.*S1);
X2 = fftn(x.*S2);
X3 = fftn(x.*S3);
X4 = fftn(x.*S4);

Z1 = X1(1:R:end,:);
Z2 = X2(1:R:end,:);
Z3 = X3(1:R:end,:);
Z4 = X4(1:R:end,:);

z1 = real(ifftn(Z1));
z2 = real(ifftn(Z2));
z3 = real(ifftn(Z3));
z4 = real(ifftn(Z4));

y1 = zeros(size(x));
y1(1:Ls/Ncoils*5/6,:) = z1(1:Ls/Ncoils*5/6,:);
y2 = zeros(size(x));
y2(Ly/Ncoils-Ls/Ncoils/6+1:Ly/Ncoils+Ls/Ncoils*5/6,:) = [z2(end-5*Ls/Ncoils/6+1:end,:);z2(1:Ls/Ncoils/6,:)];
y3 = zeros(size(x));
y3(Ly/Ncoils*2-Ls/Ncoils/6+1:Ly/Ncoils*2+5/6*Ls/Ncoils,:) = [z3((end-Ls/Ncoils/6+1):end,:);z3(1:5*Ls/Ncoils/6,:)];
y4 = zeros(size(x));
y4(end-5/6*Ls/Ncoils+1:end,:) = z4(end-5/6*Ls/Ncoils+1:end,:);
g = (y1+y2+y3+y4)./Stot;

figure,
imshow(uint8(x));
title('Original sample');

figure,
subplot(2,2,1);
imshow(uint8(x.*S1));
subplot(2,2,2);
imshow(uint8(x.*S2));
subplot(2,2,3);
imshow(uint8(x.*S3));
subplot(2,2,4);
imshow(uint8(x.*S4));
title('View of coils');

figure,
subplot(2,2,1);
imshow(uint8(z1));
subplot(2,2,2);
imshow(uint8(z2));
subplot(2,2,3);
imshow(uint8(z3));
subplot(2,2,4);
imshow(uint8(z4));
title('Undersampled coil images (R=2)');

figure,
subplot(2,2,1);
imshow(uint8(y1));
subplot(2,2,2);
imshow(uint8(y2));
subplot(2,2,3);
imshow(uint8(y3));
subplot(2,2,4);
imshow(uint8(y4));
title('Reconstructed contribution of each coil');

figure,
image(g-x);colorbar;
title('Error');

%% SENSE
clear all;
close all;
%x = LoadSignal('MRI', 256);
x = double(imread('genou.jpg'));

Ncoils = 6;
R = 4;
[Ly,Lx] = size(x);

% Generating the sensitivities
S = cell(1,Ncoils);
for c = 1:Ncoils
    S{c}(:,:) = exp(-((-Ly/2:Ly/2-1)-Ly/2+Ly/Ncoils/2+(c-1)*Ly/Ncoils).^2/2/(8*Ly))'*ones(1,Lx);
end

% Generating acquisition data
d = cell(1,Ncoils);
figure;
for c = 1:Ncoils
    subplot(2,3,c);
    imshow(uint8(x.*S{c}));
    tmp = fftn(x.*S{c});
    d{c} = ifftn(tmp(1:R:end,:));
end
title('Undersampled coil images');

% Creating the sensitivities matrices
C = cell(1,Ly/R);
for y = 1:Ly/R
    C{y} = zeros(Ncoils,R);
    for r = 1:R
        tmp = zeros(1,Ly);
        tmp(y+(r-1)*Ly/R)=1;
        Tmp = cell(1,Ncoils);
        for c = 1:Ncoils
            Tmp{c} = fft(S{c}(:,1)'.*tmp);
            Tmpi = Tmp{c}(1:R:end);
            tmpi = ifft(Tmpi);
            C{y}(c,r) = tmpi(y);
        end
    end
end

% Reconstructing data
rho = zeros(size(x));
for y = 1:Ly/R
    I = zeros(Ncoils,Lx);
    for c = 1:Ncoils
        I(c,:) = d{c}(y,:);
    end
    tmp = pinv(C{y})*I;
    for r = 1:R
        rho(y+(r-1)*Ly/R,:)=tmp(r,:);
    end
end
figure,
imshow(uint8(rho));
title('Reconstruction');

%% SMASH
clear all;
close all;
%x = LoadSignal('MRI', 256);
x = double(imread('genou.jpg'));

Ncoils = 6;
R = 2;
[Ly,Lx] = size(x);

% Generating the sensitivities
S = cell(1,Ncoils);
for c = 1:Ncoils
    S{c}(:,:) = exp(-((-Ly/2:Ly/2-1)-Ly/2+Ly/Ncoils/2+(c-1)*Ly/Ncoils).^2/2/(8*Ly))'*ones(1,Lx);
end

% Generating acquisition data
d = cell(1,Ncoils);
D = cell(1,Ncoils);
figure;
for c = 1:Ncoils
    subplot(2,4,c);
    imshow(uint8(x.*S{c}));
    tmp = fftn(x.*S{c});
    D{c} = tmp(1:R:end,:);
    d{c} = ifftn(D{c});
end
title('Coil full FOV images');

% Computing the weights
M = zeros(Ncoils,Ly);
for c = 1:Ncoils
    M(c,:)=S{c}(:,1);
end
M = M';

figure;
leg = [];
e = cell(1,R);
for r = 1:R
    e{r} = exp(-i*(r-1)*2*pi/Ly*(0:Ly-1));
    subplot(2,R,r);
    hold on;
    plot(real(e{r}),'b-');
    title(['harmonic ' num2str(r-1) ', real part']);
    hold off;
    subplot(2,R,r+R);
    hold on;
    plot(imag(e{r}),'c-');
    title(['harmonic ' num2str(r-1) ', imaginary part']);
    hold off;
end

n = cell(1,R);
iM = pinv(M);
for r = 1:R
    n{r} = iM*e{r}.';
    subplot(2,R,r);
    hold on;
    plot(real(M*n{r}),'r-');axis([0 Ly -1.1 1.1]);
    legend('Harmonic','Approximation',0);
    hold off;
    subplot(2,R,r+R);
    hold on;
    plot(imag(M*n{r}),'m-');axis([0 Ly -1.1 1.1]);
    legend('Harmonic','Approximation',0);
    hold off;
end

% Reconstruction
I = zeros(size(x));
for y = 1:Ly/R
    for r = 1:R
        tmp = zeros(1,Lx);
        for c = 1:Ncoils
            tmp = tmp+n{r}(c)*D{c}(y,:);
        end
        I((y-1)*R+r,:) = tmp;
    end
end

figure;
subplot(1,3,1);
imshow(uint8(x));
title('Original Image');
subplot(1,3,2);
imshow(uint8(real(ifftn(I))));
title('Reconstruction');
subplot(1,3,3);
imagesc((real(ifftn(I)))-x);axis square;colorbar;
title('Reconstruction error');